#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#ifndef WIN32
#include <sys/time.h>
#include <GL/glut.h>
#ifndef GLX_ARB_get_proc_address
#ifdef __cplusplus
	extern "C" void (*glXGetProcAddressARB(const GLubyte *));
#else
	extern void (*glXGetProcAddressARB(const GLubyte *));
#endif
#endif
#define glXGetProcAddress glXGetProcAddressARB
#else
#include <windows.h>
#define glXGetProcAddress wglGetProcAddress
#include "GL/gl.h"
#include "GL/glext.h"
#include <GL/glut.h>
#endif

#define VTEX_DIM_X 512 
#define VTEX_DIM_Y 512 


/* Helper functions and macros */

#define CHECKGLEXT(extstr, name) \
(strstr(extstr, #name) ? 1 : (fprintf(stderr, "Required OpenGL extension %s not found!\n", #name), 0))

#define GETGLEXTPROC(type, name) \
((name ## _  = (type) glXGetProcAddress ( (GLubyte*) #name )), \
( name ## _ ? 1 : (fprintf(stderr, "Function %s was not found\n",#name), 0 )))

#define SDECLGLEXTPROC(type, name) \
static type name ## _;

#define DECLGLEXTPROC(type, name) \
type name ## _;


#ifdef __cplusplus
	#define SCDEFGLEXTPROC(class, type, name) \
	type class :: name ## _ = NULL;
#endif

#define SDEFGLEXTPROC(type, name) \
type name ## _ = NULL;

#define CHECKERR(a) do { int i = glGetError (); if (i != GL_NO_ERROR) printf ("   *** %s: 0x%x = %s\n", a, i, gluErrorString (i)); } while (0)

PFNGLACTIVETEXTUREARBPROC glActiveTextureARB_;

PFNGLGENFRAMEBUFFERSEXTPROC	glGenFramebuffersEXT_;
PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebufferEXT_;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT_;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT_;
PFNGLGENRENDERBUFFERSEXTPROC glGenRenderbuffersEXT_;
PFNGLBINDRENDERBUFFEREXTPROC glBindRenderbufferEXT_;
PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT_;
PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT_;

int g_width = 128, g_height = 128;

GLuint g_vertexTexture;
GLuint g_imageTexture;
GLuint g_fb;
GLuint g_depth;

static const struct {
		GLuint format;
		char *name;
		int isIntFormat;
	} texformats[] = {
		{GL_ALPHA4, "GL_ALPHA4", 0},
		{GL_ALPHA8, "GL_ALPHA8", 0},
		{GL_ALPHA12, "GL_ALPHA12", 0},
		{GL_ALPHA16, "GL_ALPHA16", 0},
		
		{GL_DEPTH_COMPONENT16, "GL_DEPTH_COMPONENT16", 0},
		{GL_DEPTH_COMPONENT24, "GL_DEPTH_COMPONENT24", 0},
		{GL_DEPTH_COMPONENT32, "GL_DEPTH_COMPONENT32", 0},
		
		{GL_LUMINANCE4, "GL_LUMINANCE4", 0},
		{GL_LUMINANCE8, "GL_LUMINANCE8", 0},
		{GL_LUMINANCE12, "GL_LUMINANCE12", 0},
		{GL_LUMINANCE16, "GL_LUMINANCE16", 0},
		
		{GL_LUMINANCE4_ALPHA4, "GL_LUMINANCE4_ALPHA4", 0},
		{GL_LUMINANCE6_ALPHA2, "GL_LUMINANCE6_ALPHA2", 0},
		{GL_LUMINANCE8_ALPHA8, "GL_LUMINANCE8_ALPHA8", 0},
		{GL_LUMINANCE12_ALPHA4, "GL_LUMINANCE12_ALPHA4", 0},
		{GL_LUMINANCE12_ALPHA12, "GL_LUMINANCE12_ALPHA12", 0},
		{GL_LUMINANCE16_ALPHA16, "GL_LUMINANCE16_ALPHA16", 0},
		
		{GL_INTENSITY4, "GL_INTENSITY4", 0},
		{GL_INTENSITY8, "GL_INTENSITY8", 0},
		{GL_INTENSITY12, "GL_INTENSITY12", 0},
		{GL_INTENSITY16, "GL_INTENSITY16", 0},
		
		{GL_R3_G3_B2, "GL_R3_G3_B2", 0},
		{GL_RGB4, "GL_RGB4", 0},
		{GL_RGB5, "GL_RGB5", 0},
		{GL_RGB8, "GL_RGB8", 0},
		{GL_RGB10, "GL_RGB10", 0},
		{GL_RGB12, "GL_RGB12", 0},
		{GL_RGB16, "GL_RGB16", 0},
		
		{GL_RGBA2, "GL_RGBA2", 0},
		{GL_RGBA4, "GL_RGBA4", 0},
		{GL_RGB5_A1, "GL_RGB5_A1", 0},
		{GL_RGBA8, "GL_RGBA8", 0},
		{GL_RGB10_A2, "GL_RGB10_A2", 0},
		{GL_RGBA12, "GL_RGBA12", 0},
		{GL_RGBA16, "GL_RGBA16", 0},
		
		{GL_HILO8_NV, "GL_HILO8_NV", 0},
		{GL_SIGNED_HILO8_NV, "GL_SIGNED_HILO8_NV", 0},
		{GL_HILO16_NV, "GL_HILO16_NV", 0},
		{GL_SIGNED_HILO16_NV, "GL_SIGNED_HILO16_NV", 0},
		
		{GL_DSDT8_NV, "GL_DSDT8_NV", 0},
		{GL_DSDT8_MAG8_NV, "GL_DSDT8_MAG8_NV", 0},
		{GL_DSDT8_MAG8_INTENSITY8_NV, "GL_DSDT8_MAG8_INTENSITY8_NV", 0},
		
		{GL_FLOAT_R16_NV, "GL_FLOAT_R16_NV", 0},
		{GL_FLOAT_R32_NV, "GL_FLOAT_R32_NV", 0},
		{GL_FLOAT_RG16_NV, "GL_FLOAT_RG16_NV", 0},
		{GL_FLOAT_RG32_NV, "GL_FLOAT_RG32_NV", 0},
		{GL_FLOAT_RGB16_NV, "GL_FLOAT_RGB16_NV", 0},
		{GL_FLOAT_RGB32_NV, "GL_FLOAT_RGB32_NV", 0},
		{GL_FLOAT_RGBA16_NV, "GL_FLOAT_RGBA16_NV", 0},
		{GL_FLOAT_RGBA32_NV, "GL_FLOAT_RGBA32_NV", 0},
		
		{GL_ALPHA_FLOAT16_ATI, "GL_ALPHA_FLOAT16_ATI", 0},
		{GL_ALPHA_FLOAT32_ATI, "GL_ALPHA_FLOAT32_ATI", 0},
		{GL_INTENSITY_FLOAT16_ATI, "GL_INTENSITY_FLOAT16_ATI", 0},
		{GL_INTENSITY_FLOAT32_ATI, "GL_INTENSITY_FLOAT32_ATI", 0},
		{GL_LUMINANCE_FLOAT16_ATI, "GL_LUMINANCE_FLOAT16_ATI", 0},
		{GL_LUMINANCE_FLOAT32_ATI, "GL_LUMINANCE_FLOAT32_ATI", 0},
		{GL_LUMINANCE_ALPHA_FLOAT16_ATI, "GL_LUMINANCE_ALPHA_FLOAT16_ATI", 0},
		{GL_LUMINANCE_ALPHA_FLOAT32_ATI, "GL_LUMINANCE_ALPHA_FLOAT32_ATI", 0},
		{GL_RGB_FLOAT16_ATI, "GL_RGB_FLOAT16_ATI", 0},
		{GL_RGB_FLOAT32_ATI, "GL_RGB_FLOAT32_ATI", 0},
		{GL_RGBA_FLOAT16_ATI, "GL_RGBA_FLOAT16_ATI", 0},
		{GL_RGBA_FLOAT32_ATI, "GL_RGBA_FLOAT32_ATI", 0},
		
		{GL_ALPHA16F_ARB, "GL_ALPHA16F_ARB", 0},
		{GL_ALPHA32F_ARB, "GL_ALPHA32F_ARB", 0},
		{GL_INTENSITY16F_ARB, "GL_INTENSITY16F_ARB", 0},
		{GL_INTENSITY32F_ARB, "GL_INTENSITY32F_ARB", 0},
		{GL_LUMINANCE16F_ARB, "GL_LUMINANCE16F_ARB", 0},
		{GL_LUMINANCE32F_ARB, "GL_LUMINANCE32F_ARB", 0},
		{GL_LUMINANCE_ALPHA16F_ARB, "GL_LUMINANCE_ALPHA16F_ARB", 0},
		{GL_LUMINANCE_ALPHA32F_ARB, "GL_LUMINANCE_ALPHA32F_ARB", 0},
		{GL_RGB16F_ARB, "GL_RGB16F_ARB", 0},
		{GL_RGB32F_ARB, "GL_RGB32F_ARB", 0},
		{GL_RGBA16F_ARB, "GL_RGBA16F_ARB", 0},
		{GL_RGBA32F_ARB, "GL_RGBA32F_ARB", 0},

		{GL_RGBA32UI_EXT,"GL_RGBA32UI_EXT", 1},
		{GL_RGB32UI_EXT,"GL_RGB32UI_EXT", 1},
		{GL_ALPHA32UI_EXT,"GL_ALPHA32UI_EXT", 1},
		{GL_INTENSITY32UI_EXT,"GL_INTENSITY32UI_EXT", 1},
		{GL_LUMINANCE32UI_EXT,"GL_LUMINANCE32UI_EXT", 1},
		{GL_LUMINANCE_ALPHA32UI_EXT,"GL_LUMINANCE_ALPHA32UI_EXT", 1},

		{GL_RGBA16UI_EXT,"GL_RGBA16UI_EXT", 1},
		{GL_RGB16UI_EXT,"GL_RGB16UI_EXT", 1},
		{GL_ALPHA16UI_EXT,"GL_ALPHA16UI_EXT", 1},
		{GL_INTENSITY16UI_EXT,"GL_INTENSITY16UI_EXT", 1},
		{GL_LUMINANCE16UI_EXT,"GL_LUMINANCE16UI_EXT", 1},
		{GL_LUMINANCE_ALPHA16UI_EXT,"GL_LUMINANCE_ALPHA16UI_EXT", 1},

		{GL_RGBA8UI_EXT,"GL_RGBA8UI_EXT", 1},
		{GL_RGB8UI_EXT,"GL_RGB8UI_EXT", 1},
		{GL_ALPHA8UI_EXT,"GL_ALPHA8UI_EXT", 1},
		{GL_INTENSITY8UI_EXT,"GL_INTENSITY8UI_EXT", 1},
		{GL_LUMINANCE8UI_EXT,"GL_LUMINANCE8UI_EXT", 1},
		{GL_LUMINANCE_ALPHA8UI_EXT,"GL_LUMINANCE_ALPHA8UI_EXT", 1},

		{GL_RGBA32I_EXT,"GL_RGBA32I_EXT", 1},
		{GL_RGB32I_EXT,"GL_RGB32I_EXT", 1},
		{GL_ALPHA32I_EXT,"GL_ALPHA32I_EXT", 1},
		{GL_INTENSITY32I_EXT,"GL_INTENSITY32I_EXT", 1},
		{GL_LUMINANCE32I_EXT,"GL_LUMINANCE32I_EXT", 1},
		{GL_LUMINANCE_ALPHA32I_EXT,"GL_LUMINANCE_ALPHA32I_EXT", 1},

		{GL_RGBA16I_EXT,"GL_RGBA16I_EXT", 1},
		{GL_RGB16I_EXT,"GL_RGB16I_EXT", 1},
		{GL_ALPHA16I_EXT,"GL_ALPHA16I_EXT", 1},
		{GL_INTENSITY16I_EXT,"GL_INTENSITY16I_EXT", 1},
		{GL_LUMINANCE16I_EXT,"GL_LUMINANCE16I_EXT", 1},
		{GL_LUMINANCE_ALPHA16I_EXT,"GL_LUMINANCE_ALPHA16I_EXT", 1},

		{GL_RGBA8I_EXT,"GL_RGBA8I_EXT", 1},
		{GL_RGB8I_EXT,"GL_RGB8I_EXT", 1},
		{GL_ALPHA8I_EXT,"GL_ALPHA8I_EXT", 1},
		{GL_INTENSITY8I_EXT,"GL_INTENSITY8I_EXT", 1},
		{GL_LUMINANCE8I_EXT,"GL_LUMINANCE8I_EXT", 1},
		{GL_LUMINANCE_ALPHA8I_EXT,"GL_LUMINANCE_ALPHA8I_EXT"}
	};


int  getExtensions();

void testFBO() {
	GLenum fbStatus;
	int i, count;
	int maxAttachments, maxSize, maxDrawBuffers;

	glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &maxAttachments);
	glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &maxSize);
	glGetIntegerv(GL_MAX_DRAW_BUFFERS, &maxDrawBuffers);

	fprintf(stderr, "General Properties:\n");
	fprintf(stderr, "\tMAX_COLOR_ATTACHMENTS=%i\n", maxAttachments);
	fprintf(stderr, "\tMAX_RENDERBUFFER_SIZE=%i\n", maxSize);
	fprintf(stderr, "\tMAX_DRAW_BUFFERS=%i\n", maxDrawBuffers);

	glGenTextures(1, &g_imageTexture);
	glGenFramebuffersEXT_(1, &g_fb);

	glActiveTextureARB_(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, g_imageTexture);

	count = 0;

	fprintf(stderr, "Supported Formats:\n");
	for (i = 0; i < sizeof(texformats) / sizeof(texformats[0]); i++) {
		glTexImage2D(GL_TEXTURE_2D, 0, texformats[i].format, VTEX_DIM_X, VTEX_DIM_Y, 0,
		             texformats[i].isIntFormat ? GL_RGBA_INTEGER_EXT : GL_RGBA,
		             GL_BYTE, NULL);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_REPEAT);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		glBindFramebufferEXT_(GL_FRAMEBUFFER_EXT, g_fb);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		glFramebufferTexture2DEXT_(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
								GL_TEXTURE_2D, g_imageTexture, 0);
		if (glGetError() != GL_NO_ERROR) {
			continue;
		}
		fbStatus = glCheckFramebufferStatusEXT_(GL_FRAMEBUFFER_EXT);    
		switch(fbStatus) {
			case GL_FRAMEBUFFER_COMPLETE_EXT:
				fprintf(stderr, "\t%s... ok\n", texformats[i].name);
				count++;
				break;
			case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
				break;
			default: 
				break;
		}
	}

	glBindFramebufferEXT_(GL_FRAMEBUFFER_EXT, 0);
	fprintf(stderr, "%d formats supported\n", count);
	exit(0);
}

int init() 
{
	if (!getExtensions()) {
		exit(1);
	}

	testFBO();

	return 1;
}



/**
   Check for OpenGL extensions and get function pointers if available
*/
int getExtensions()
{
	char *glExtString = (char*)glGetString(GL_EXTENSIONS);

	/*fprintf(stderr, "%s\n", glExtString);*/

	if (!CHECKGLEXT(glExtString, GL_ARB_multitexture)){
		return 0;
	}
	GETGLEXTPROC(PFNGLACTIVETEXTUREARBPROC, glActiveTextureARB);

	if (!CHECKGLEXT(glExtString, GL_EXT_framebuffer_object)){
		return 0;
	}
	GETGLEXTPROC(PFNGLGENFRAMEBUFFERSEXTPROC, glGenFramebuffersEXT);
	GETGLEXTPROC(PFNGLBINDFRAMEBUFFEREXTPROC, glBindFramebufferEXT);
	GETGLEXTPROC(PFNGLFRAMEBUFFERTEXTURE2DEXTPROC, glFramebufferTexture2DEXT);
	GETGLEXTPROC(PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC, glCheckFramebufferStatusEXT);
	GETGLEXTPROC(PFNGLGENRENDERBUFFERSEXTPROC, glGenRenderbuffersEXT);
	GETGLEXTPROC(PFNGLBINDRENDERBUFFEREXTPROC, glBindRenderbufferEXT);
	GETGLEXTPROC(PFNGLRENDERBUFFERSTORAGEEXTPROC, glRenderbufferStorageEXT);
	GETGLEXTPROC(PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC, glFramebufferRenderbufferEXT);

	return 1;
}

int main(int argc, char *argv[]) 
{
	glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_ALPHA | GLUT_DEPTH);
	glutInitWindowSize(g_width, g_height);
	glutCreateWindow(argv[0]);

	if (!init()) {
		exit(1);
	}

	glutMainLoop();
    
	return 0;
}

